/*
 * sourcebase.h
 * Copyright (C) Biswapesh Chattopadhyay 2003
 * This file is the main high-level interface to sourcebase.
 */

#ifndef _SOURCEBASE_H
#define _SOURCEBASE_H

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#ifdef __cplusplus
extern "C"
{
#endif

/* Symbol types recognized by sourcebase. To ensure binary compatibility,
 * please append to this list only. Note that some of them are redundant
 * and should be removed before binary compatibility is enforced.
 */
typedef enum _SBSymType {
	SB_TYPE_DEF,
	SB_CLASS_DEF,
	SB_STRUCT_DEF,
	SB_MBR_FUNC_DEF,
	SB_MBR_VAR_DEF,
	SB_ENUM_DEF,
	SB_CONS_DEF,
	SB_MACRO_DEF,
	SB_FUNC_DEF,
	SB_SUBR_DEF,
	SB_GLOB_VAR_DEF,
	SB_COMMON_DEF,
	SB_COMMON_MBR_VAR_DEF,
	SB_CLASS_INHERIT,
	SB_MBR_FUNC_DCL,
	SB_FUNC_DCL,
	SB_ENUM_CONST_DEF,
	SB_UNION_DEF,
	SB_FRIEND_DCL,
	SB_NAMESPACE_DEF, /* Not handled in the parsers ? */
	SB_EXCEPTION_DEF, /* Not handled in the parsers ? */
	SB_TEMPLATE_DEF,
	SB_LOCAL_VAR_DEF,
	SB_VAR_DCL,
	SB_INCLUDE_DEF,
	SB_COMMENT_DEF, /* Not handled in the parsers ? */
	SB_MAX_DEF
} SBSymType;

/* Human readable names of the symbols */
extern const char *sb_sym_type_names[SB_MAX_DEF];

/* Cross reference values. Some of them are not being currently used but should
 * probably be kept for future parsers.
 */
typedef enum _SBXrefType {
	SB_REF_TO_TYPEDEF = SB_TYPE_DEF,
	SB_REF_TO_DEFINE = SB_MACRO_DEF,
	SB_REF_TO_ENUM = SB_ENUM_CONST_DEF,
	SB_REF_TO_STRUCT = SB_STRUCT_DEF,
	SB_REF_TO_UNION = SB_UNION_DEF,
	SB_REF_TO_CLASS = SB_CLASS_DEF,
	SB_REF_TO_FUNCTION = SB_FUNC_DEF,
	SB_REF_TO_MBR_FUNC = SB_MBR_FUNC_DEF,
	SB_REF_TO_MBR_VA = SB_MBR_VAR_DEF,
	SB_REF_TO_COMM_VAR = SB_COMMON_MBR_VAR_DEF,
	SB_REF_TO_CONSTANT = SB_CONS_DEF,
	SB_REF_TO_SUBROUTINE = SB_SUBR_DEF,
	SB_REF_TO_GLOB_VAR = SB_GLOB_VAR_DEF,
	SB_REF_TO_LOCAL_VAR = SB_LOCAL_VAR_DEF,
	SB_REF_TO_TEMPLATE = SB_TEMPLATE_DEF,
	SB_REF_TO_NAMESPACE = SB_NAMESPACE_DEF,
	SB_REF_TO_EXCEPTION = SB_EXCEPTION_DEF,
	SB_REF_TO_LABEL = SB_SUBR_DEF,
	SB_REF_UNDEFINED = SB_MAX_DEF
} SBXrefType;

/* Scope of reference */
typedef enum _SBRefScope {
	SB_REF_SCOPE_LOCAL,
	SB_REF_SCOPE_GLOBAL
} SBRefScope;

/* Variable references */
typedef enum _SBVarRefType {
	SB_REF_READ,
	SB_REF_WRITE,
	SB_REF_PASS,
	SB_REF_UNUSED
} SBVarRefType;

/* Symbol attributes. */
typedef enum _SBSymAttr {
	SB_PRIVATE = 0x000001,
	SB_PROTECTED = 0x000002,
	SB_PUBLIC = 0x000004,
	SB_STATIC = 0x000008,
	SB_VIRTUAL = 0x001000,
	SB_ABSTRACT = 0x000010,
	SB_FINAL = 0x000020,
	SB_NATIVE = 0x000040,
	SB_SYNCHRONIZED = 0x000080,
	SB_VOLATILE = 0x000100,
	SB_TRANSIENT = 0x000200,
	SB_INTERFACE = 0x000400,
	SB_IMPLEMENTS = 0x000800,
	SB_INLINE = 0x002000,
	SB_CONSTRUCTOR = 0x004000,
	SB_DESTRUCTOR = SB_CONSTRUCTOR,
	SB_PUREVIRTUAL = (0x008000 | SB_VIRTUAL),
	SB_OVERRIDE = 0x20000,
	SB_OVERLOADED = 0x40000
} SBSymAttr;

/* Parser attributes. Please append to this list only for ABI compatibility */
typedef enum _SBParserAttr {
	SB_PARSE_XREF = 1, /* Supports xref generation */
	SB_PARSE_COMMENT = 2, /* Supports comment generation */
	SB_PARSE_NOCASESENSE = 4 /* Language is not case sensitive */
} SBParserAttr;

/* Action to be taken for the word - used by the C preprocessor */
typedef enum _SBMWordType {
	SB_WORD_DEFINE = 1,
	SB_WORD_UNDEF = 2,
	SB_WORD_REPLACE = 3,
	SB_WORD_NONE = 4
} SBMWordType;

/* Flags set by options passed to the parser. Use sb_ref_set and
 * sb_pref_get to get/set these preferences at the rpoject level.
 */
typedef enum _SBOptions
{
	SB_OPT_CACHESIZE, /* Cache size to be used by the database */
	SB_OPT_CASE_SENSITIVE, /* Whether parsing should be case-sensitive */
	SB_OPT_COMMENTS, /* Whether the comment database is to be built */
	SB_OPT_DIALECT, /* Language dialect to use */
	SB_OPT_DBPREFIX, /* Common prefix to use for the database file(s) */
	SB_OPT_DROP_USR_HEADERS, /* Whether system headers are to be dropped */
	SB_OPT_PARSER, /* List of parsers available */
	SB_OPT_INCLUDE_LIST, /* List of directories to scan for include files */
	SB_OPT_LOCAL_VARS, /* Whether to populate the local variables database */
	SB_OPT_TREAT_AS_CPLUSPLUS, /* Whether to parse as C++ */
	SB_OPT_XREF, /* Whether cross-reference database is to be built */
	SB_OPT_XREF_FILENAME
} SBOptions;

/* Type of directory */
typedef enum _SBDirType {
	SB_DIR_SOURCE = 1,
	SB_DIR_HEADER = 2,
	SB_DIR_BOTH = 3,
} SBDirType;

/**
 * Type of component.
 * This is the field comp_type in table PARSERS in the DB.
*/
typedef enum _SBComponentType	{
	SB_COMPONENT_EXECUTABLE = 1,
	SB_COMPONENT_SHAREDLIB = 2
} SBComponentType;
/**
 * Type of directory entry.
 */
typedef enum _SBDirEntryType	{
	SB_DIRENTRY_FILE = 1,
	SB_DIRENTRY_DIRECTORY = 2
} SBDirEntryType;

/* Query results should be returned by sourcebase backends in this format.
 * All data items are strings. NULL is an allowed value. The first ncols
 * strings should be the column names, so the actual data starts from
 * the ncol-th position. The total number of strings = ncols * (nrows + 1).
 * The dataset is 0-index based (start with row=0, col=0). */
typedef struct _SBDataset {
	int nrows; /* No. of rows fetched by this query. */
	int ncols; /* No. of columns per row in this resultset */
	char **data; /* The actual data as an array of char pointers. */
} SBDataset;

#define sb_data(D, row, col) (D)->data[((row) * (D)->ncols) + col]

/**
 * Describes a file in a directory
 */
typedef struct _DBFilesTableElem	{
	int file_id; /** Unique identifier for this file, taken from the DB */
	char *file_name; /** Name of the file (no path) */
	char *file_dir; /** Directory where the file resides (full path) */
	char *file_path; /** name + dir - saved for convenience (full path) */
	int parser_id; /** ID of the parser to be used with this file, taken from the DB */
	time_t parse_time; /** Unix timestamp indicating when this file was last parsed */
	int file_attr; /** Extra attributes of the file (bitwise OR-ed values) */
} DBFilesTableElem;

typedef struct _DBDirectoriesTableElem	{
	char *dir_path; /* The directory */
	int dir_type; /* Source (1) or include (2) or both (3)*/
	char *dir_order; /* Order in which the directories should be scanned for files */
	int dir_attr; /* Extra attributes saved as an integer of bitwise OR-ed values */
} DBDirectoriesTableElem;

/**
 * Directory entry.
 */
typedef struct _SBDirEntry	{
	SBDirEntryType EntryType; /** One of SBDirEntryType. If SB_DIRENTRY_FILE then FileElem is valid and DirElem == NULL */
	DBFilesTableElem *FileElem; /** Pointer to file description */
	DBDirectoriesTableElem *DirElem; /** Pointer to directory description */
} SBDirEntry;

typedef struct _DBParsersTableElem	{
	int parser_id; /** Parser code */
	char *parser_name; /** A human readable name for the parser (eg 'python') */
	char *comp_name; /** Filename of the component (e.g. /usr/lib/sourcebase/parsers/libpybrowser.so */
	int comp_type; /** Component type, i.e executable or shared lib */
	int parser_attr; /** Attributes indicating features this parser supports */
} DBParsersTableElem;

/*
 * Execute a query on the underlying database and return the result as
 * a dataset. Note that is can be only supported by backends that support
 * SQL.
 */
SBDataset *sb_execute_query (const char *sql);

void sb_dataset_free (SBDataset *ds);

/* Initialize the sourcebase library. Call this once at the start
 * @return TRUE on success, FALSE on failure
 */
int sb_init (void);

/* DB manipulation functions */

/* Initialize a project - some default values will be populated automatically
 * by the DB backend when a project is initialized. However, the project is
 * not automatically scanned at this stage. You should use the function
 * sn_project_autoscan to automatically create the project symbol and xref
 * database once you've customized the database. Note that only one project
 * is allowed to be in open state at a time and the function will fail if you
 * haven't closed the previous project using sn_project_close before calling
 * this function.
 * @param dir Top level directory of the project.
 * @return 0 on success, -1 on failure.
 */
int sb_project_init (const char *dir);

int sb_project_autoscan (int recurse);

/*
 * Add all available parsers to the project database. This funntion should
 * be used only by backend providers.
 * @return 0 on success, -1 on failure.
 */
int sb_add_parsers(void);

/* Closes the current project and frees all information associated with it.
 * @return 0 on success, -1 on failure.
 */
int sb_project_close (void);

/* Gets the top directory of the current open project. Note that NULL indicates
 * that no project is currently open - so you can also use this function to test
 * whether a project is open or not. The returned string is a pointer to an
 * internal bugger and *MUST NOT BE free()D*
 */
const char *sb_project_get_dir (void);

/*
 * Put a symbol into the project database. This should be called by the parser
 * for all symbols it encounters.
 * @param type The symbol type (SBSymType)
 * @param scope Scope of the symbol
 * @return 0 on success, -1 on failure
 */
int sb_put_symbol (int type,char *scope,char *sym_name,char *file,
    int start_lineno,int start_colpos,int end_lineno,int end_colpos,
    unsigned long attr,char *ret,char *arg_types,char *args);

/* Gets a symbol from the project database.
 * @param class_name Name of the class
 */
int sb_get_symbol (char *class_name,char *exp_scope,char *name,
    char *arg_types, char *scope,char *ret_type, char *ret_define, int exact);

/* Gets a class, struct or typedef information from the project database.
 * name Name of the symbol.
 * @return Symbol type ?
 */
int sb_get_compound (char *name,char *origin);

/* Put a cross reference record into the XREF table.
 * @param type Type of xref (SBXRefType)
 * @return 0 on success, -1 on failure
 */
int sb_put_xref (int type,int scope_type,int scope_lev,char *fnc_cls,
    char *fnc,char *fnc_arg_types,char *scope,char *what,char *arg_types,
    char *file,int lineno,int acc);

/* Put a file into the project database for parsing. This function will
 * also call the relevant function to parse this file if required. It
 * will also take care of clearing the symbol database of all symbols
 * and xrefs for this file if required.
 * @param file_name Name of the file with path. Can be relative to the
 * current dir, top project directory or an absolute path. The function
 * will take care of locating the appropriate file if required.
 * @param parser The name of the parser to use with this file. This is
 * optional. If NULL is passed, an appropriate parser will be selected
 * automatically based on existing data and/or file extension.
 * @return 0 on success, -1 on failure.
 */
int sb_put_file (char *file_name, char *parser);
int sb_put_file_full (DBFilesTableElem elem);

int sb_put_dir_full (DBDirectoriesTableElem elem);
/* Populate the comments table.
 * @param classn The name of the class
 * @return 0 on success, -1 on failure
 */
int sb_put_comment (char *classn,char *func,char *filename,char *comment,
    int beg_line,int beg_char);

/**
 * \fn sb_find_file
 * \brief Get the absolute path of the file.
 * You can pass an absolute path and it will test its existance.
 * If a relative path it will try:
 *	-# To find it in the project database, if the project is open
 *	-# To find it relative to the project root dir if the project is open
 *	-# To find it relative to the current directory
 *	-# To find it in system include directory. 
 * \sa includepath
 * \param file_name The name of the file to search
 * \param path Should be an empty buffer of size PATH_MAX. Will contain the path to the file on return.
 * \return 0 on success, -1 on failure.
 */
int sb_find_file (char *file_name, char *path);

/*
 * Get the parser to be used for parsing the file.
 * @param file_name The name of the file.
 * @return The name of the parser. Note that this is a constant string
 * and should not be free()d. If no parser is found, NULL is returned.
 */
const char *sb_get_parser_for_file(const char *file_name);

/*
 * Parse a file with the given parser.
 * @param file_name The name of the file. If file_name is NULL, all
 * files in the database which have been modified since last parse
 * are parsed.
 * @param parser The parser to use for parsing the files. If NULL is
 * passed, a correct parser is chosen automatically. Note that if
 * file_name is NULL, specifying the parser has no effect.
 * @param force If set to 1, the passed file_name will be parsed even
 * if it has not been modified since the last parse.
 * @return 0 on success, -1 on failure.
 */
int sb_parse_file(const char *file_name, const char *parser, int force);

/*
 * Set a preference for the project. See SBOptions for list of preferences.
 * @param pref Id of the pref
 * @value A string with the value of the pref
 * @return 0 on success, -1 on failure
 */
int sb_pref_set(int pref, const char *value);

/*
 * Same as sb_pref_set, except that the value to set is an integer.
 * @param pref Id of the pref
 * @value A string with the value of the pref
 * @return 0 on success, -1 on failure
 */
int sb_pref_set_int(int pref, int value);

/*
 * Get the value of a preference. The returned value is a newly allocated
 * string which should be free()d when no longer required.
 * @param pref The Id of the pref whose value is required.
 * @param value Will contain the string corresponding to the value
 * @return 0 on success, NULL on failure.
 */
int sb_pref_get(int pref, char **value);

/*
 * Same as sb_pref_get except that the resutned value is an integer.
 * @param pref The Id of the pref whose value is required.
 * @param value Will contain the string corresponding to the value
 * @return 0 on success, NULL on failure.
 */
int sb_pref_get_int(int pref, int *value);

#ifdef __cplusplus
}
#endif

#endif /* _SOURCEBASE_H */
