/*
Copyright (c) 2000, Red Hat, Inc.
Copyright (c) 2003, Biswapesh Chattopadhyay
*/

%{
/*
 * pybrowser.l
 *
 * Written by Thomas Heller <thomas.heller@ion-tof.com>
 * Modified by Biswapesh Chattopadhyay <biswapesh_chatterjee@tcscal.co.in>
 *
 * Description:
 * A lex specification for a scanner to parse python source code files
 * and store relevant information into the Source-Navigator project database.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <string.h>

#include "sourcebase-parser.h"

/* Name of the parser */
const char parser_name[] = "python";

/* File extensions handled by the parser */
const char parser_extns[] = "py PY Py";

static void var_or_const(int);

static struct stack *scope_stack;

static void push(int, int, char *);
static void pop(int);
static char *top(int *);
static int calc_indent (char*, char *);
static void baseclasses (char *, char *);
static void parse_triple(int);

/* XXX Is this correct? */
#undef YY_INPUT
#define YY_INPUT(buf,r,ms) (r = sb_encoded_input(buf, ms))

int
yywrap()
{
   return(1);
} 
  
%}

%%

\"\"\" {
    sb_advance_column (3);
    parse_triple((int)'"');
}

\'\'\' {
    sb_advance_column (3);
    parse_triple((int)'\'');
}

^[ \t]*def[ \t]+[a-zA-Z_][a-zA-Z_0-9]*[ \t]*\( {
    int ofs;
    char *p;
    int indent;
    char *classname;
    int code;
    int context;
    /* point just before the '(', then go backwards to find and mark
       the end of the name */
    p = yytext + yyleng - 1;
    while ((p[-1] == ' ') || (p[-1] == '\t'))
	--p;
    *p = '\0';
    /* now go backwards to find the beginning */
    p = yytext + yyleng-1;
    while ((p > yytext) && (p[-1] != ' ') && (p[-1] != '\t'))
	--p;
    ofs = p - yytext;

    indent = calc_indent (yytext, p);
    pop(indent);
    classname = top(&context);
    if (classname) {
	switch (context) {
	case SB_CLASS_DEF:
	    code = SB_MBR_FUNC_DEF;
	    break;
	default:
	    code = SB_FUNC_DEF;
	    classname = NULL;
	    break;
	}
    } else
	code = SB_FUNC_DEF;

    push(indent, code, p);

    sb_put_symbol(code, classname, p, sb_current_file(),
		     sb_line(), sb_column()+ofs, sb_line(), 
		     sb_column()+ofs + strlen(p),
		     0, NULL, NULL, NULL);
}

^[ \t]*class[ \t]+[a-zA-Z_][a-zA-Z_0-9]*[ \t]*(\([^)]*\))?[ \t]*: {
    int ofs;
    char *p, *super;
    int indent;
    char *end;
    yytext[yyleng] = '\0';

    p = yytext;
    while (isspace (*p))
        ++p;
    indent = calc_indent (yytext, p);

    p += strlen ("class");
    while (isspace (*p))
        ++p;
    ofs = p - yytext;
    end = p;
    while (*end && (isalnum (*end) || (*end == '_')))
        ++end;
    *end = '\0';

    pop(indent);
    push(indent, SB_CLASS_DEF, p);
    sb_insert_symbol(SB_CLASS_DEF, NULL, p, sb_current_file(),
		     sb_line(), sb_column()+ofs, sb_line(), 
		     sb_column()+ofs + strlen(p),
		     0, NULL, NULL, NULL, NULL,
		     sb_line(), sb_column()+ofs, sb_line(),
		     sb_column()+ofs + strlen(p));
    ++end;
    super = end;
    while ((end < yytext+yyleng) && *end) {
	if (*end == '\n') {
            sb_advance_line();
            sb_reset_column();
	}
        ++end;
    }
    baseclasses (p, super);
}

^[A-Z_][A-Z0-9_]*[ \t]*= { var_or_const (SB_CONS_DEF); }
  
^[a-zA-Z_][a-zA-Z0-9_]*[ \t]*= { var_or_const (SB_GLOB_VAR_DEF); }
  
\n {
   sb_advance_line();
   sb_reset_column();
}

. { sb_advance_column(yyleng); }

%%
struct stack {
    int indent;
    int code;
    char *name;
    struct stack *prev;
};

static struct stack *scope_stack = NULL;

static int calc_indent(char *start, char *stop)
{
    int indent = 0;
    char *cp = start;
    while (cp < stop) {
	switch (*cp++) {
	case ' ':
	    ++indent;
	    break;
	case '\t':
	    indent = (indent / 8) * 8 + 8;
	    break;
	default:
	    return indent;
	}
    }
    return indent;
}

static char *top(int *code)
{
    if (scope_stack) {
	*code = scope_stack->code;
	return scope_stack->name;
    }
    return NULL;
}

static void push(int indent, int code, char *name)
{
    struct stack *node;

    node = (struct stack *)malloc(sizeof (struct stack));
    node->indent = indent;
    node->code = code;
    node->name = strdup(name);
    node->prev = scope_stack;
    scope_stack = node;
}

static void pop(int indent)
{
    while (scope_stack && (scope_stack->indent >= indent)) {
	scope_stack = scope_stack->prev;
    }
}

static void baseclasses (char *classname, char *rest)
{
    /* We enter here with something like " ( Tester ):" */
    char *cp = rest;
    char *base;
	char *tmp;

    while (base = strtok_r (cp, "():, \t\r\n", &tmp)) {
	if (strchr (base, '.'))
	    base = strrchr (base, '.') + 1;
	cp = NULL;
	/* I don't really know why we need a filename here, the
	   base class may be defined in a totally different file */
	sb_insert_symbol(SB_CLASS_INHERIT, classname, base,
			 sb_current_file(),
			 sb_line(), sb_column(), sb_line(), 
			 sb_column() + strlen(base),
			 0, NULL, NULL, NULL, NULL,
			 sb_line(), sb_column(), sb_line(),
			 sb_column() + strlen(base));
    }
}

static void var_or_const (int code)
{
    char *p;
    /* point just before the '=', then go backwards to find and mark
       the end of identifier */
    p = yytext + yyleng - 1;
    while ((p > yytext) && (p[-1] == ' ') || (p[-1] == '\t'))
        --p;
    *p = '\0';
    sb_insert_symbol(code, NULL, yytext, sb_current_file(),
		     sb_line(), sb_column(), sb_line(), 
		     sb_column() + strlen(yytext),
		     0, NULL, NULL, NULL, NULL,
		     sb_line(), sb_column(), sb_line(),
		     sb_column() + strlen(yytext));
}

static void parse_triple(int term)
{
    int c;
    int count;
    while (1) {
	c = input();
	sb_advance_column (1);
	if (c == term) {
	    ++count;
	    if (count == 3)
		return;
	} else {
	    count = 0;
	    switch (c) {
	    case '\\':
		c = input();
		if (c <= 0)
		    return;
		if (c != '\n') {
		    sb_advance_column (1);
		    c = -1;
		} else {
		    sb_advance_line ();
		}
		break;

	    case '\n':
		sb_advance_line ();
		sb_reset_column ();
		break;

	    case '\r':
		sb_reset_column ();
		break;

	    case EOF:
		return;
	    }
	}
    }
}

void reset()
{
   sb_reset_line();
   sb_reset_column();
   sb_reset_encoding();
}

/* Function that takes a file name and parses it */
int
parse_file(const char *file_name)
{
   return sb_main(file_name, parser_name, &yyin, yylex, reset);
}
