/*
 * sourcebase-parser.c
 *
 * Copyright (C) 1997 Cygnus Solutions, Inc.
 * Copyright (C) 2003 Biswapesh Chattopadhyay
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <glib.h>

#include "sourcebase-parser.h"

static long line = 1;
static long column = 0;

static long savedLine;
static long savedColumn;

static char *current_parser = NULL;
static char *current_filename = NULL;

static int case_sensitive = 1;
static int comments = 0;
static int dialect = 0;
static int drop_usr_headers = 0;
static int local_vars = 0;
static int treat_as_cplusplus = 0;

static start_of_file = 1;
static FILE * yyin;

static void
sb_set_parser(const char *parser)
{
	g_return_if_fail(parser);
	if (current_parser)
		g_free(current_parser);
	current_parser = g_strdup(parser);
}

static void
sb_set_filename(const char *file_name)
{
	g_return_if_fail(current_filename);
	if (current_filename)
		g_free(current_filename);
	current_filename = g_strdup(file_name);
}

int
sb_parse_main(const char *file_name, const char *parser, FILE ** lexstream
  , sb_lexer_fn_t lexer, sb_lexer_reset_fn_t reset)
{
	sb_set_parser(parser);
	if (*lexstream)
		*lexstream = freopen(file_name, "rb", *lexstream);
	else
	*lexstream = fopen(file_name, "rb");
	if (NULL == lexstream) {
		perror(file_name);
		return -1;
	}
	sb_set_filename(file_name);
	reset();
	return lexer();
}

/*
 * Set the current column to a new position.
 */
void
sb_set_column(long c)
{
   column = c;
}

/*
 * Reset the column position.
 */
void
sb_reset_column()
{
   sb_set_column(0);
}

/*
 * Advance the column position by `num' positions.
 */
void
sb_advance_column(int num)
{
  column += num;
}

/*
 * Set the current line to a new position.
 */
void
sb_set_line(long l)
{
  line = l;
}

/*
 * Reset the line position.
 */
void
sb_reset_line()
{
  sb_set_line(1);
}

/*
 * Advance to the next line position.
 */
void
sb_advance_line()
{
  line++;
} 

/*
 * Retrieve the current line position.
 */
long
sb_line()
{
  return line;
}

/*
 * Retrieve the current column position.
 */
long
sb_column()
{
  return column;
}

/*
 * Save the current line position on a stack.
 */
void
sb_push_line()
{
  savedLine = sb_line();
}

/*
 * Pop the stored line position off the stack.
 */
long
sb_pop_line()
{
  long result = savedLine;
  savedLine = -1;
  return result;
} 

/*
 * Push the current column position onto the stack.
 */
void
sb_push_column()
{
  savedColumn = sb_column();
}

/*
 * Pop the stored column position off the stack.
 */
long
sb_pop_column()
{
  long result = savedColumn;

  savedColumn = -1;
  return result;
}

/*
 * Count the number of line and column advancements in a null-terminated
 * buffer.
 */
void 
sb_count_chars(char *buf, int length)
{
  char *p;
  int i;

  for (p = buf, i = length; i > 0; i--, p--)
  {
    if (*p == '\n')
    {
      sb_advance_line();
      sb_reset_column();
    }
    else
    {
      sb_advance_column(1);
    }
  }
}

/*
 * Return the filename of the current source file.
 */
char *
sb_current_file()
{
  return current_filename;
}

int
sb_encoded_input(char *buf, int max_size)
{
	return fread(buf, sizeof(char), max_size, yyin);
}

void
sb_reset_encoding()
{
  start_of_file = 1;
}
