/*
** sourcebase.c: Common implementation functions
** Copyright (C) 2003 Biswapesh Chattopadhyay
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <gmodule.h>

#include "sourcebase.h"
#include "backend.h"
#include "utils.h"

#ifndef PACKAGE_BACKEND_DIR
#define PACKAGE_BACKEND_DIR NULL
#endif
#ifndef PACKAGE_PARSER_DIR
#define PACKAGE_PARSER_DIR "/usr/local/lib/sourcebase/parsers"
#endif
#define PARSER_FILE_PATTERN "*." G_MODULE_SUFFIX

const char *includepath[] = {
	"/usr/include",
	"/usr/local/include",
	NULL
};
/* Top directory of the current project */
static char *top_project_dir = NULL;
static GList *parser_list = NULL;
const char *sb_sym_type_names[] = {
	"typedef",
	"class",
	"member function definition",
	"member variable",
	"enum",
	"constant",
	"macro",
	"function definition",
	"subroutine",
	"global variable",
	"common block",
	"common member variable",
	"inheritance",
	"member function declaration",
	"function declaration",
	"enum value",
	"union",
	"friend",
	"namespace", /* Not handled in the parsers ? */
	"exception", /* Not handled in the parsers ? */
	"local variable",
	"variable",
	"include",
	"comment", /* Not handled in the parsers ? */
	NULL
};


typedef int (*parse_fn_t) (const char *file_name);

/* Structure storing information about parsers */
typedef struct _SBParserImpl {
	const char *name; /* Name of the parser */
	const char *file_name; /* File name of the parser library/executable */
	gboolean is_executable; /* Whether the parser is a stand-alone exe */
	GModule *module; /* The loaded module */
	parse_fn_t parse_fn; /* Function to parse a file */
	char *extns; /* Extensions supported by this parser */
	int attrs;
} SBParserImpl;

/* SQLite is the default (currently, only) backend */

static SBParserImpl *get_parser_impl_by_name (char *name)	{
	g_return_val_if_fail (parser_list, NULL);
	GList *list = parser_list;
	SBParserImpl *impl = NULL;

	while (list)	{
		impl = (SBParserImpl *) list->data;
		if (!impl)
			continue;
		if (0 == g_ascii_strcasecmp ((const gchar*) name, (const gchar *) impl->name))
			return impl;
	}
	return NULL;
}
/* Load all available parsers and their default properties */
static gboolean sb_load_parsers(void)
{
	GDir *p_dir;
	const gchar *name;

	if (NULL == (p_dir = g_dir_open(PACKAGE_PARSER_DIR, 0, NULL)))
		return FALSE;
	while (NULL != (name = g_dir_read_name(p_dir))) {
		if (g_pattern_match_simple(PARSER_FILE_PATTERN, name)) {
			char *parser_file = g_build_filename (PACKAGE_PARSER_DIR, name);
			SBParserImpl *pi = g_new0(SBParserImpl, 1);
			pi->module = g_module_open(parser_file, G_MODULE_BIND_LAZY);
			if (NULL == pi->module) {
				g_warning("Unable to open parser module %s", parser_file);
				g_free(parser_file);
				g_free(pi);
				continue;
			}
			pi->file_name = g_strdup(parser_file);
			g_module_symbol(pi->module, "parser_name", (gpointer *) &(pi->name));
			g_module_symbol(pi->module, "parse_file", (gpointer *) &(pi->parse_fn));
			g_module_symbol(pi->module, "parser_extns", (gpointer *) &(pi->extns));
			g_module_symbol(pi->module, "parser_attrs", (gpointer *) &(pi->attrs));
			parser_list = g_list_prepend(parser_list, pi);
			g_free(parser_file);
		}
	}
	g_dir_close(p_dir);
	return TRUE;
}

int sb_init (void)	{
	if (TRUE != sb_load_parsers())
		return -1;
	return 0;
}

int sb_add_parsers(void)
{
	GList *t;
	SBParserImpl *p;
	for (t = parser_list; t; t = g_list_next(t)) {
		p = (SBParserImpl *) t->data;
        if (0 != add_parser(p->name, p->file_name, p->extns, p->attrs))
			return -1;
	}
	return 0;
}

int sb_project_init(const char *dir)
{
	g_return_val_if_fail(dir, -1);
	if (NULL != top_project_dir) {
		g_warning("Project %s already open!", top_project_dir);
		return -1;
	}
	if (! g_file_test(dir, G_FILE_TEST_IS_DIR)) {
		g_warning("%s: No such directory", dir);
		return -1;
	}
	top_project_dir = g_strdup(dir);
	return init_db(dir);
}


int sb_project_autoscan (int recurse)	{
	g_return_val_if_fail (top_project_dir, -1);

	gchar *extn = NULL;
	gchar *parser_id = NULL;
	
	GHashTable *parsed_exns = get_parsed_extns_from_db ();
	g_return_val_if_fail (parsed_exns, -1);
	//~ These are existing files in the project DB, as DBFilesTableElem *
	GHashTable *existing_files = get_project_files_from_db ();
	g_return_val_if_fail (existing_files, -1);
	
	GHashTable *parsers = get_parsers_from_db ();
	g_return_val_if_fail (parsers, -1);

	GList *parse_tree = traverse_path (top_project_dir, (recurse == 0) ? FALSE : TRUE);
	g_return_val_if_fail (parse_tree, -1);
	
	GList *elem = parse_tree;
	GList *update_files = NULL;
	SBDirEntry *data = NULL;
	DBFilesTableElem *FileElem = NULL;
	DBDirectoriesTableElem *DirElem = NULL;
	DBFilesTableElem *ExistingElem = NULL;
	DBDirectoriesTableElem *ExistingDirElem = NULL;
	DBParsersTableElem *ParserElem = NULL;
	SBParserImpl *parser = NULL;
	struct stat statbuf;

	//~ Keep all the directories, and files for which there's a parser
	while (elem)	{
		data = (SBDirEntry *)elem->data;
		switch (data->EntryType)	{
			case (SB_DIRENTRY_FILE):
				FileElem = data->FileElem;
				extn = get_file_extension (FileElem->file_name);
				//~ g_debug_message ("Checking if file %s (%s) is to be parsed.", FileElem->file_path, extn);
				parser_id = (gchar *)g_hash_table_lookup (parsed_exns, extn);
				if (!parser_id)	{
					//~ g_debug_message ("No, extension %s not supported", extn);
					g_free (data);
				}	else	{
					FileElem->parser_id = atoi(parser_id);
					update_files = g_list_append (update_files, data);
					//~ g_debug_message ("Yes, handled by parser id %d", atoi(parser_id));
				}
				break;
			case (SB_DIRENTRY_DIRECTORY):
				//~ g_debug_message ("Adding directory %s", data->DirElem->dir_path);
				update_files = g_list_append (update_files, data);
				break;
			default:
				break;
		}
		elem = g_list_next (elem);
	}
	update_files = g_list_first (update_files);
	elem = update_files;

	//~ Identify which files/dirs are new and which files have been modified.
	//~ All updated/new files and new dirs go to update_files
	while (elem)	{
		data = (SBDirEntry *)elem->data;
		switch (data->EntryType)	{
			case (SB_DIRENTRY_FILE):
				FileElem = data->FileElem;
				//~ absolute file path is the key of the hash table
				ExistingElem = g_hash_table_lookup (existing_files, FileElem->file_path);
				//~ File exists
				if (ExistingElem)	{
					if (0 != stat (FileElem->file_path, &statbuf))
						//~ TODO: Cleanup
						return -1;
					if (FileElem->parse_time < statbuf.st_mtime)	{
						update_files = g_list_append (update_files, FileElem);
						//~ g_debug_message ("File %s already in DB and to be updated", FileElem->file_path);
					}
				}	else	{
					//~ g_debug_message ("File %s is new", FileElem->file_path);
					update_files = g_list_append (update_files, FileElem);
				}
				break;
			case (SB_DIRENTRY_DIRECTORY):
				DirElem = data->DirElem;
				ExistingDirElem = g_hash_table_lookup (existing_files, DirElem->dir_path);
				if ((!ExistingDirElem) || (ExistingDirElem->dir_type != DirElem->dir_type)||
						(ExistingDirElem->dir_order != DirElem->dir_order) || (ExistingDirElem->dir_attr != DirElem->dir_attr))	{
					//~ g_debug_message ("Updating dir %s", DirElem->dir_path);
					update_files = g_list_append (update_files, DirElem);
				}
				break;
			default:
				break;
		}
		elem = g_list_next (elem);
	}
	update_files = g_list_first (update_files);
	elem = update_files;
	while (elem)	{
		//~ TODO: call parse function and sb_put_file_full
		data = (SBDirEntry *) elem->data;
		switch (data->EntryType)	{
			case (SB_DIRENTRY_FILE):
				FileElem = data->FileElem;
				ParserElem = g_hash_table_lookup (parsers, &FileElem->parser_id);
				if (!ParserElem)
					break;
				//~ g_debug_message ("Found parser %s (ID %d) for file %s", ParserElem->parser_name, ParserElem->parser_id, FileElem->file_name);
				parser = get_parser_impl_by_name (ParserElem->parser_name);
				if (!parser)
					break;
				//~ TODO: Got here :-)
				parser->parse_fn (FileElem->file_name);
				break;
			case (SB_DIRENTRY_DIRECTORY):
				DirElem = data->DirElem;
				//~ if (0 != sb_put_dir_full (*DirElem))	{
					//~ TODO: cleanup
					//~ return -1;
				//~ }
				break;
			default:
				break;
		}
		elem = g_list_next (elem);
	}

	g_list_free (parse_tree);
	g_list_free (update_files);
	g_hash_table_destroy (parsed_exns);
	g_hash_table_destroy (parsers);
	g_hash_table_destroy (existing_files);
	return 0;
}

const char *sb_project_get_dir(void) {
	return top_project_dir;
}

int sb_project_close(void)
{
	if (NULL == top_project_dir) {
		g_warning("No project open!");
		return -1;
	}
	g_free(top_project_dir);
	top_project_dir = NULL;
    return close_db();
}


int sb_put_xref (int type,int scope_type,int scope_lev,char *fnc_cls,
    char *fnc,char *fnc_arg_types,char *scope,char *what,char *arg_types,
    char *file,int lineno,int acc)
{
	int xref_is_on = 0;
	if (0 != sb_pref_get_int(SB_OPT_XREF_FILENAME, &xref_is_on)) {
		g_warning("Unable to get xref preference");
		return -1;
	}
	if (xref_is_on)
		return put_xref (type, scope_type, scope_lev, fnc_cls, fnc
		  , fnc_arg_types, scope, what, arg_types, file, lineno, acc);
	else
		return 0;
}

int sb_find_file(char *file_name, char *path)
{
	g_return_val_if_fail(file_name && ('\0' != file_name[0]), -1);
	g_return_val_if_fail(path, -1);

	if (g_path_is_absolute (file_name))	{
		//~ Absolute file path.
		if (g_file_test(file_name, G_FILE_TEST_IS_REGULAR))	{
			g_snprintf (path, PATH_MAX, file_name);
			return 0;
		} else	{
			return -1;
		}
	}
	if (top_project_dir) { // Project is open
		// Ask the backend
		int status = find_file_in_db (file_name, path);
		if ((0 == status) && (g_path_is_absolute (path)))
			return 0;
		else if (0 == status)	{
			// See if the file is relative to the top project directory
			char *newpath = g_build_filename (top_project_dir, file_name);
			if (g_file_test(newpath, G_FILE_TEST_IS_REGULAR))	{
				g_snprintf (path, PATH_MAX, "%s", newpath);
				g_free (newpath);
				return 0;
			}
			g_free (newpath);
		}
	}
	char *curdir = g_get_current_dir ();
	// Perhaps path is relative to current dir rather than project dir
	if (strcmp (curdir, top_project_dir) != 0)	{
		char *newpath = g_build_filename (curdir, file_name);
		if (g_file_test(newpath, G_FILE_TEST_IS_REGULAR))	{
			g_snprintf (path, PATH_MAX, "%s", newpath);
			g_free (curdir);
			g_free (newpath);
			return 0;
		}
		g_free (newpath);
	}
	g_free (curdir);
	/* Might be a system header */
	int i;
	char *try = NULL;
	for (i = 0; includepath[i]; i++)	{
		if (g_file_test (includepath[i], G_FILE_TEST_IS_DIR))	{
			try = g_build_filename (includepath[i], file_name);
			if (g_file_test(try, G_FILE_TEST_IS_REGULAR))	{
				g_snprintf (path, PATH_MAX, "%s", try);
				g_free  (try);
				return 0;
			}
			
		}
	}
	g_free  (try);

	/* No dice :-( */
	return -1;
}

int sb_pref_set_int(int pref, int value)
{
	char *val_str;
	int status;
	val_str = g_strdup_printf("%d", value);
	status = sb_pref_set (pref, val_str);
	g_free(val_str);
	return status;
}

int sb_pref_get_int(int pref, int *value)
{
	char *val_str = NULL;
	int status;
	status = sb_pref_get (pref, &val_str);
	if (0 == status)
		*value = atoi(val_str);
	else
		*value = 0;
	return status;
}
