/**
 * @file sqlite-backend.c
 * A backend implementation for sourcebase using SQLite database.
 * @author Biswapesh Chattopadhyay
 * @date 2003
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <sqlite.h>

#include "sourcebase.h"
#include "utils.h"
#include "backend.h"

#define DBNAME ".sourcebase.sqlite-db"
#define CREATEDB_FILE_NAME PACKAGE_MODULE_DIR "/sqlite-createdb.sql"

/**
 * \var db
 * \brief Static variable that identifies DB instance
 */
static sqlite *db = NULL;
/**
 * \var top_project_dir
 * \brief Absolute path of project root
 */
static char *top_project_dir = NULL;
static char *sqlite_err = NULL;
static char last_file[PATH_MAX] = "";
static int last_file_id = -1;

/**
 * \var register_symbol_sql
 * \brief Prototype for inserting a symbol into the DB
 */
static const char register_symbol_sql[] = "INSERT INTO TYPES ( \
  sym_type, type_name) VALUES (%d, %Q);";

/**
 * \fn register_symbol_types
 * \brief For each of the symbol types register into the DB
 * \sa _SBSymType;
 * \return int
 * 	- 0 if OK
 * 	- -1 if SQLite error
 */
static int register_symbol_types(void)
{
	int i;
	for (i=0; sb_sym_type_names[i] != NULL; ++i) {
		if (SQLITE_OK != sqlite_exec_printf(db, register_symbol_sql, NULL,
			NULL, &sqlite_err, i, sb_sym_type_names[i])) {
			g_warning("Unable to register symbol type %s: %s",
			  sb_sym_type_names[i], sqlite_err ? sqlite_err :
			  "Unknown sqlite error");
			return -1;
		}
	}
	return 0;
}

/**
 * Initializes the db and saves the instance as static member.
 * 
 * @param dir Optional, in what directory you want to create the DB. NULL means current dir.
 * @sa db
 * @todo init_db should return different values, depending on what went wrong. Perhaps an enum would do
 * @return int
 *	- 0 if OK
 *	- -1 if fail
 */
int init_db (const char *dir)
{
	char *db_file;
	char *errmsg = NULL;
	gboolean db_exists = FALSE;
	
	if (NULL != db) {
		g_warning("DB already initialized");
		return -1;
	}
	if (!g_file_test(CREATEDB_FILE_NAME, G_FILE_TEST_IS_REGULAR)) {
		g_warning("Table creation script %s  missing", CREATEDB_FILE_NAME);
		return -1;
	}
	if (top_project_dir)
		g_free(top_project_dir);
	
	if (NULL != dir)
		top_project_dir = g_strdup(dir);
	else
		top_project_dir = g_get_current_dir();
	
	db_file = g_build_filename(top_project_dir, DBNAME, NULL);
	if (g_file_test(db_file, G_FILE_TEST_IS_REGULAR))
		db_exists = TRUE;
	
	db = sqlite_open(db_file, 0, &errmsg);
	if (NULL == db) {
		if (NULL != errmsg)
			g_warning("%s", errmsg);
		else
			g_warning("Unknown error trying to create DB file %s", db_file);
		g_free (db_file);
		return -1;
	}
	if (!db_exists) { /* New database - populate it */
		GError *err = NULL;
		gchar *sql_stmt = NULL;
		if (!g_file_get_contents(CREATEDB_FILE_NAME, &sql_stmt, NULL, &err)) {
			g_warning("Unable to cread file %s: %s", CREATEDB_FILE_NAME,
			  err ? err->message : "");
			g_free (db_file);
			return -1;
		}
		if (SQLITE_OK != sqlite_exec(db, sql_stmt, NULL, NULL, &sqlite_err)) {
			g_warning("Unable to create project tables: %s", sqlite_err ?
			  sqlite_err : "Unknown SQLIte error");
			g_free (db_file);
			return -1;
		}
		/* Create a list of parsers */
		if (0 != sb_add_parsers()) {
			g_warning("Unable to register parsers");
			g_free (db_file);
			return -1;
		}
		/* Create a list of symbol types */
		if (0 != register_symbol_types()) {
			g_free (db_file);
			return -1;
		}
	}
	g_free (db_file);
	return 0;
}

/**
 * \var add_parser_sql
 * \brief Prototype of SQL code to insert a parser into the database
 */
static const char add_parser_sql[] = "INSERT into PARSERS (\
  parser_id, parser_name, comp_name, comp_type, parser_attr) values (\
  NULL, %Q, %Q, 2, %d);";
/**
 * \var add_parser_extn_sql
 * \brief Prototype of SQL code to insert a supported fileextension for a specified parser
 */
static const char add_parser_extn_sql[] = "INSERT INTO PARSER_EXTNS (\
  parser_id, extn) values (%d, %Q);";

/**
 * \fn add_parser
 * \brief Adds a parser to the database.
 * \param name A human readable name for the parser
 * \param file_name Name of the component
 * \param extns Space separated list of wildcards that parser supports, e.g. "*.c *.cpp"
 * \param attrs Attributes indicating features this parser supports 
 * \return
 */
int add_parser(const char *name, const char *file_name, const char *extns
  , int attrs)
{
	char **extn_array;
	char **extn;
	int parser_id;
	g_return_val_if_fail(name, -1);
	g_return_val_if_fail(file_name, -1);
	g_return_val_if_fail(extns, -1);
	g_return_val_if_fail(db, -1);
	if (SQLITE_OK != sqlite_exec_printf(db, add_parser_sql, NULL, NULL,
		&sqlite_err, name, file_name, attrs)) {
			g_warning("Unable to register parser %s: %s", name, sqlite_err ?
			sqlite_err : "Unknown sqlite error");
			return -1;
	}
	parser_id = sqlite_last_insert_rowid(db);
	extn_array = g_strsplit(extns, " ", 50);
	if (NULL == extn_array) {
		/* A parser whcih supports no language */
		g_warning("Created a parser with no language support");
		return 0;
	}
	for (extn = extn_array; *extn; ++ extn) {
		if (SQLITE_OK != sqlite_exec_printf(db, add_parser_extn_sql, NULL,
			NULL, &sqlite_err, parser_id, *extn)) {
			g_warning("Error inserting parser extension: %s", sqlite_err ?
				sqlite_err : "Unknown sqlite error");
			g_strfreev(extn_array);
			return -1;
		}
	}
	g_strfreev(extn_array);
	return 0;
}


static const char get_files_sql[] = "SELECT * FROM FILES;";
GHashTable *get_project_files_from_db (void)	{
	g_return_val_if_fail (db, NULL);

	gint nrows = 0;
	gint ncols = 0;
	gchar **files = NULL;
	gchar **file = NULL;
	GHashTable *FileHash = NULL;
	DBFilesTableElem *Elem = NULL;
	gint row = 1;
	
	if (SQLITE_OK != sqlite_get_table(db, get_files_sql, &files, &nrows, &ncols,
		&sqlite_err)) {
		g_warning("Unable to execute query: %s", sqlite_err ? sqlite_err :
		  "Unknown sqlite error");
		return NULL;
	}

	FileHash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
	for (; row <= nrows; row++)	{
		file = files+(row*ncols);
		Elem = g_new0 (DBFilesTableElem, 1);
		Elem->file_id = atoi(file[0]);
		Elem->file_name = g_strdup(file[1]);
		Elem->file_dir = g_strdup(file[2]);
		Elem->file_path = g_strdup(file[3]);
		Elem->parser_id = atoi(file[4]);
		Elem->parse_time = atoi(file[5]);
		Elem->file_attr = atoi(file[6]);
		//~ g_debug_message ("File Path %s", Elem->file_path);
		g_hash_table_insert (FileHash, Elem->file_path, Elem);
	}
	sqlite_free_table (files);
	return FileHash;
}

static const char get_parsed_extns_sql[] = "SELECT parser_id, extn FROM PARSER_EXTNS;";
GHashTable *get_parsed_extns_from_db (void)	{
	g_return_val_if_fail(db, NULL);
	
	gint nrows = 0;
	gint ncols = 0;
	gchar **extns = NULL;
	gchar **extn = NULL;
	GHashTable *ExtHash = NULL;
	gint row = 1;
	
	if (SQLITE_OK != sqlite_get_table(db, get_parsed_extns_sql, &extns, &nrows, &ncols,
		&sqlite_err)) {
		g_warning("Unable to execute query: %s", sqlite_err ? sqlite_err :
		  "Unknown sqlite error");
		return NULL;
	}

	ExtHash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
	for (; row <= nrows; row++)	{
		extn = extns+(row*ncols);
		//~ g_debug_message("Extension %s handled by %s", extn[1], extn[0]);
		g_hash_table_insert (ExtHash, g_strdup(extn[1]), g_strdup(extn[0]));
	}
	sqlite_free_table (extns);
	return ExtHash;
}


static const char get_parsers_sql[] = "SELECT * FROM PARSERS;";
GHashTable *get_parsers_from_db (void)	{
	g_return_val_if_fail(db, NULL);
	
	gint nrows = 0;
	gint ncols = 0;
	gchar **parsers = NULL;
	gchar **parser = NULL;
	GHashTable *ParserHash = NULL;
	gint row = 1;
	DBParsersTableElem *Elem = NULL;

	if (SQLITE_OK != sqlite_get_table(db, get_parsers_sql, &parsers, &nrows, &ncols,
		&sqlite_err)) {
		g_warning("Unable to read parsers from DB: %s", sqlite_err ? sqlite_err :
		  "Unknown sqlite error");
		return NULL;
	}

	//~ ParserHash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
	ParserHash = g_hash_table_new_full (g_int_hash, g_int_equal, NULL, g_free);
	for (; row <= nrows; row++)	{
		parser = parsers+(row*ncols);
		Elem = g_new0 (DBParsersTableElem, 1);
		Elem->parser_id = atoi(parser[0]);
		Elem->parser_name = g_strdup(parser[1]);
		Elem->comp_name = g_strdup(parser[2]);
		Elem->comp_type = atoi(parser[3]);
		Elem->parser_attr = atoi(parser[4]);
		//~ g_hash_table_insert (ParserHash, g_strdup_printf ("%d", Elem->parser_id), Elem);
		g_hash_table_insert (ParserHash, &Elem->parser_id, Elem);
		//~ g_debug_message ("Parser %s id %d", Elem->parser_name, Elem->parser_id);
	}
	sqlite_free_table (parsers);
	return ParserHash;
}

/**
 * Execute an SQL query on this backend instance.
 * @param sql SQL code to execute
 * @return NULL if any error occurred, or a g_new allocated result set.
 */
SBDataset *sb_execute_query (const char *sql)
{
	int nrows = 0;
	int ncols = 0;
	char **data = NULL;

	g_return_val_if_fail(db, NULL);
	g_return_val_if_fail(sql, NULL);
	if (SQLITE_OK != sqlite_get_table(db, sql, &data, &nrows, &ncols,
		&sqlite_err)) {
		g_warning("Unable to execute query: %s", sqlite_err ? sqlite_err :
		  "Unknown sqlite error");
		return NULL;
	} else {
		SBDataset *ds = g_new (SBDataset, 1);
		ds->nrows = nrows;
		ds->ncols = ncols;
		ds->data = data;
		return ds;
	}
}

/**
 * \fn sb_dataset_free
 * \brief Free a result set obtained from execute_query
 */
void sb_dataset_free(SBDataset *ds)
{
	if (ds) {
		sqlite_free_table(ds->data);
		g_free (ds);
	}
}

/*
 * Locate a file in the project database
 * @param file_name Name of the file
 * @param path Contains the file path on return (absolute/relative to top dir)
 * @return file_id if found in FILES table, 0 if found otherwise, -1 is not found
 */
static char file_id_sql[] = "SELECT file_id from FILES where file_path = %Q;";
static char dir_sql[] = "SELECT dir_path from DIRECTORIES order by \
  dir_order, dir_type;";
static int get_file_id(const char *file_name, char *path)
{
	SBDataset ds;
	int file_id = -1;
	int i;

	g_return_val_if_fail(db, -1);
	g_return_val_if_fail(file_name, -1);
	g_return_val_if_fail(path, -1);
	if (SQLITE_OK != sqlite_get_table_printf(db, file_id_sql, &(ds.data),
		&(ds.nrows), &(ds.ncols), &sqlite_err, file_name)) {
		g_warning("Error trying to get file id: %s", sqlite_err ? sqlite_err :
			"Unknown sqlite error");
		return -1;
	}
	if (ds.nrows > 0)
		file_id = atoi(ds.data[1]);
	sqlite_free_table(ds.data);
	return file_id;
	/* No go, let's try the project directories */
/*	if (SQLITE_OK != sqlite_get_table(db, dir_sql, &(ds.data), &(ds.nrows),
		&(ds.ncols), &sqlite_err)) {
		g_warning("Error getting directories: %s", sqlite_err ? sqlite_err :
			"Unknown sqlite error");
		return -1;
	}
	for (i=1; i < ds.nrows; ++i) {
		//snprintf(path, "%s/%s", ds.data[i], file_name);
		sprintf(path, "%s/%s", ds.data[i], file_name);
		if (g_file_test(path, G_FILE_TEST_EXISTS)) {
			return 0;
		}
	}*/
	return -1;
}

static int get_parser_id(const char *file_name, const char *parser)
{
	SBDataset ds;
	int parser_id = -1;

	if (NULL == file_name && NULL == parser) {
		g_warning("Both parser and file_name cannot be NULL");
		return -1;
	}
	if (file_name) {
		/* Check if there is a parser specified for this file */
		if (SQLITE_OK != sqlite_get_table_printf(db, "SELECT parser_id FROM FILES \
			where file_path = %Q;", &(ds.data), &(ds.nrows), &(ds.ncols),
			&sqlite_err, file_name)) {
			g_warning("Error getting parser id for file %s: %s", file_name,
				sqlite_err ? sqlite_err : "Unknown sqlite error");
			return -1;
		}
		if (ds.nrows > 0) {
			parser_id = atoi(ds.data[1]);
		}
		sqlite_free_table(ds.data);
		if (parser_id != -1)
			return parser_id;
	}
	if (parser) {
		/* Get the ID for this parser */
		if (SQLITE_OK != sqlite_get_table_printf(db, "SELECT parser_id from PARSERS \
			where parser_name = %Q;", &(ds.data), &(ds.nrows), &(ds.ncols),
			&sqlite_err, parser)) {
			g_warning("Unable to get parser id for parser %s: %s", parser,
				sqlite_err ? sqlite_err : "Unknown sqlite error");
			return -1;
		}
		if (ds.nrows > 0) {
			parser_id = atoi(ds.data[1]);
		}
		sqlite_free_table(ds.data);
		if (parser_id != -1)
			return parser_id;
	}
	if (file_name) {
		if (SQLITE_OK != sqlite_get_table_printf(db, "SELECT parser_id from \
			PARSER_EXTNS where %Q GLOB extn;", &(ds.data), &(ds.nrows),
			&(ds.ncols), &sqlite_err, file_name)) {
			g_warning("Error querying parser extensions table: %s", sqlite_err ?
				sqlite_err : "Unknown sqlite error");
			return -1;
		}
		if (ds.nrows > 0) {
			parser_id = atoi(ds.data[1]);
		}
		sqlite_free_table (ds.data);
		//sqlite_free_resultset(ds.data);
	}
	return parser_id;
}

static char *put_file_sql = "INSERT INTO FILES (file_id, file_name, file_dir \
  , file_path, parser_id, parse_time, file_attr) VALUES (NULL, %Q, %Q, %Q \
  , %d, NULL, 0);";

int sb_put_file (char *file_name, char *parser)
{
	//int parser_id;
	int file_id;
	char path[PATH_MAX];

	g_return_val_if_fail(file_name, -1);
	if (-1 == (file_id = get_file_id(file_name, path))) {
		/* File does not exist. Find a suitable parser */
		int parser_id = get_parser_id(file_name, parser);
		if (0 > parser_id) {
			/* Didn't get a parser :-( */
			g_warning("No parser found for %s", file_name);
			return -1;
		}
		/* @todo Get file name and directory from path */
		if (SQLITE_OK != sqlite_exec_printf(db, put_file_sql, NULL, NULL,
			&sqlite_err, file_name, NULL, path, parser_id)) {
			g_warning("Unable to insert file %s: %s", file_name, sqlite_err ?
				sqlite_err : "Unknown sqlite error");
			return -1;
		}
	}
	return 0;
}

static char *put_file_full_sql = "INSERT INTO FILES (file_id, file_name, file_dir \
  , file_path, parser_id, parse_time, file_attr) VALUES (NULL, %Q, %Q, %Q \
  , %d, %d, %d);";
static char *update_file_sql = "UPDATE FILES SET parse_time = %d, file_attr = %d \
	WHERE file_id = %d;";

int sb_put_file_full (DBFilesTableElem elem)	{
	g_return_val_if_fail (elem.file_name, -1);
	g_return_val_if_fail (elem.file_dir, -1);
	g_return_val_if_fail (elem.file_path, -1);
	g_return_val_if_fail ((elem.parser_id != -1), -1);

	gchar path[PATH_MAX];
	gint file_id;
	
	if (-1 != (file_id = get_file_id(elem.file_path, path))) {
		g_debug_message ("File %s already exists in DB", elem.file_path);
		if (SQLITE_OK != sqlite_exec_printf(db, update_file_sql, NULL, NULL,
			&sqlite_err, elem.parse_time, elem.file_attr, elem.file_id)) {
			g_warning("Unable to update file %s: %s", elem.file_path, sqlite_err ?
				sqlite_err : "Unknown sqlite error");
			return -1;
		}
	}	else	{
		if (SQLITE_OK != sqlite_exec_printf(db, put_file_full_sql, NULL, NULL,
			&sqlite_err, elem.file_name, elem.file_dir, elem.file_path,
			elem.parser_id, elem.parse_time, elem.file_attr)) {
			g_warning("Unable to insert file %s: %s", elem.file_path, sqlite_err ?
				sqlite_err : "Unknown sqlite error");
			return -1;
		}
		return 0;
	}
	return 0;
}

//~ This just works even if an UPDATE statement would be needed. Boh...
static char *put_dir_full_sql = "INSERT INTO DIRECTORIES VALUES (%Q, %d, %Q, %d);";
int sb_put_dir_full (DBDirectoriesTableElem elem)	{
	g_return_val_if_fail (elem.dir_path, -1);
	g_return_val_if_fail (elem.dir_type, -1);

	gint result = sqlite_exec_printf(db, put_dir_full_sql, NULL, NULL, &sqlite_err,
		elem.dir_path, elem.dir_type, elem.dir_order, elem.dir_attr);
	if (SQLITE_OK != result) {
			g_warning("Unable to insert directory %s: %s(error code %d)", elem.dir_path, sqlite_err ?
			sqlite_err : "Unknown sqlite error", result);
		return -1;
	}
	return 0;
}

int put_comment (char *classn,char *func,char *filename,char *comment,
    int beg_line,int beg_char)
{
	g_warning("Function not implemented in %s:%d", __FILE__, __LINE__);
	return 0;
}

int close_db (void)
{
	if (SQLITE_OK != sqlite_exec_printf(db, "COMMIT;", NULL, NULL, &sqlite_err))
		g_warning("Unable to commit transaction: %s", sqlite_err ? sqlite_err : "Unknown sqlite error");
	sqlite_close (db);
	return 0;
}

int sb_pref_set (int pref_id, const char *value)	{
	g_warning("Function not implemented in %s:%d", __FILE__, __LINE__);
	return 0;
}
int sb_pref_get (int pref_id, char **value)	{
	g_warning("Function not implemented in %s:%d", __FILE__, __LINE__);
	return 0;
}
int find_file_in_db (const char *file_name,char *path)	{
	g_warning("Function not implemented in %s:%d", __FILE__, __LINE__);
	return 0;
}
/*
 * Given a file name, this function will try to resolve it to get
 * the actual location.
 * @param path The file name whose resolution is requested
 * @param name Will contain the name of the file on return
 * @param dir Will contain the directory of the file on return
 * @return 0 on success, -1 on failure
 */
int resolve_path(const char *file_name, char *name, char *dir)
{
	char path[PATH_MAX];
	char *current_dir = g_get_current_dir();
	gboolean just_name = (gboolean) (NULL == strchr(file_name, '/'));

	g_return_val_if_fail(path, -1);
	g_return_val_if_fail(name, -1);
	g_return_val_if_fail(dir, -1);
	if ('/' == file_name[0]) { /* Absolute path */
		snprintf(path, PATH_MAX, "%s", file_name);
		if (g_file_test(path, G_FILE_TEST_EXISTS))
			goto resolved;
		else
			goto unresolved;
	}
	else if (('.' == file_name[0]) && !just_name) {
		snprintf(path, PATH_MAX, "%s/%s", current_dir, file_name);
		if (g_file_test(path, G_FILE_TEST_EXISTS))
			goto resolved;
		else
			goto unresolved;
		
	} else {
		/* Relative to project dir ? */
		if (top_project_dir) {
			snprintf(path, PATH_MAX, "%s/%s", top_project_dir, file_name);
			if (g_file_test(path, G_FILE_TEST_EXISTS)) {
				goto resolved;
			}
		}
		/* Relative to current directory ? */
		snprintf(path, PATH_MAX, "%s/%s", current_dir, file_name);
		if (g_file_test(path, G_FILE_TEST_EXISTS)) {
			goto resolved;
		}
		/* Let's try all the project directories */
		if (0 == find_file_in_db(file_name, path))
			goto resolved;
		/* System header ? */
		snprintf(path, PATH_MAX, "/usr/include/%s", file_name);
	}
resolved:
	return 0;
unresolved:
	return -1;
}
int sb_get_symbol (char *class_name,char *exp_scope,char *name,
    char *arg_types, char *scope,char *ret_type, char *ret_define, int exact)
{
	g_warning("Function not implemented in %s:%d", __FILE__, __LINE__);
	return 0;
}

int sb_get_compound (char *name,char *origin)
{
	g_warning("Function not implemented in %s:%d", __FILE__, __LINE__);
	return 0;
}

static char put_xref_sql[] = "INSERT INTO XREF (xref_id, parent_sym, sym_name \
  , sym_type, arg_types, ref_parent_sym, ref_sym_name, ref_sym_type, \
  ref_arg_types, access_type, file_id, start_line, start_col) VALUES (\
  NULL, %Q, %Q, %d, %Q, %Q, %Q, %d, %Q, %d, %d, %d, NULL);";

int put_xref (int type,int scope_type,int scope_lev,char *fnc_cls,
    char *fnc,char *fnc_arg_types,char *scope,char *what,char *arg_types,
    char *file,int lineno,int acc)
{
	int file_id;
	char path[PATH_MAX];

	g_return_val_if_fail(db, -1);
	g_return_val_if_fail(what, -1);
	g_return_val_if_fail(file, -1);

	/* Cache the file Id for better performance */
	if (0 == strcmp(file, last_file))
		file_id = last_file_id;
	else {
		file_id = get_file_id(file, path);
		if (-1 == file_id) {
			g_warning("File %s not found in DB", file);
			return -1;
		}
		strcpy(last_file, file);
		last_file_id = file_id;
	}
	/* @todo Add validations/cleanups before inserting xref information */
	if (SQLITE_OK != sqlite_exec_printf(db, put_xref_sql, NULL, NULL
		, &sqlite_err, fnc_cls, fnc, type, fnc_arg_types, scope, what
		, scope_type, arg_types, acc, file_id, lineno)) {
		g_warning("Unable to insert xref: %s", sqlite_err ? sqlite_err :
			"Unknown sqlite error");
		return -1;
	}
	return 0;
}
static char put_sym_sql[] = "INSERT INTO SYMBOLS (\
  sym_id, sym_name, sym_type, sym_attr, parent_sym, file_id, start_line, \
  start_col, end_line, end_col, dec_type, template_class, arg_types, \
  arg_names) values (NULL, %Q, %d, %ld, %Q, %d, %d, %d, %d, %d, %Q, NULL \
  , %Q, %Q);";

int sb_put_symbol (int type,char *scope,char *sym_name,char *file,
    int start_lineno,int start_colpos,int end_lineno,int end_colpos,
    unsigned long attr,char *ret,char *arg_types,char *args)
{
	int file_id;
	char path[PATH_MAX];

	g_return_val_if_fail(sym_name, -1);
	g_return_val_if_fail(file, -1);

	/* Cache the file Id for better performance */
	if (0 == strcmp(file, last_file))
		file_id = last_file_id;
	else {
		file_id = get_file_id(file, path);
		if (-1 == file_id) {
			g_warning("File %s not found in DB", file);
			return -1;
		}
		strcpy(last_file, file);
		last_file_id = file_id;
	}
	/* @todo Add validations/cleanups before inserting xref information */
	if (SQLITE_OK != sqlite_exec_printf(db, put_sym_sql, NULL, NULL, &sqlite_err
		, sym_name, type, attr, scope, file_id, start_lineno, start_colpos
		, end_lineno, end_colpos, ret, arg_types, args)) {
		g_warning("Unable to put symbol %s: %s", sym_name, sqlite_err ?
			sqlite_err : "Unknown sqlite error");
		return -1;
	}
	return 0;
}
