/* Script to initialize a project database for the SQLITE backend */

/*
** The PARSERS table contains a list of available parsers and their properties.
*/
create table PARSERS (
  parser_id INTEGER PRIMARY KEY, /** Parser code */
  parser_name VARCHAR(64), /* A human readable name for the parser (eg 'python') */
  comp_name VARCHAR(1024), /* Filename of the component (e.g. /usr/lib/sourcebase/parsers/libpybrowser.so */
  comp_type INTEGER, /* Component type, i.e executable or shared lib */
  parser_attr INTEGER /* Attributes indicating features this parser supports */
);

/*
** The PARSER_EXTNS table lists the file extensions a parser supports
*/
create table PARSER_EXTNS (
	parser_id INTEGER,
	extn VARCHAR(8),
	PRIMARY KEY (parser_id, extn)
	/* This is not supported by SQLite */
	/* FOREIGN KEY parser_id REFERENCES PARSERS(parser_id) */
);

/*
** The TYPES table contains the various symbol types supported by the different
** parsers. Each type is a unique string (preferably small) which is used in the
** SYMBOLS table (see below) to identify the type of the SYMBOL. On creation of
** a project database, this table is automatically created. If you modify this
** table by adding/removing rows, please keep in mind that the parsers will only
** insert symbols in the SYMBOLS table which have a type entry that is present
** in this table.
*/
create table TYPES (
  sym_type VARCHAR(64) PRIMARY KEY, /* A short unique string indicating the symbol type */
  type_name VARCHAR(1024), /* A human-readable name for the type */
  description TEXT/* A full description of the type */
);

/*
** The DIRECTORIES table contains an ordered list of directories whcih are
** scanned for presence of included (dependent) files. They might contain paths
** relative to the top project directory, the current directory or absolute
** directories
*/
create table DIRECTORIES (
	dir_path VARCHAR(1024) PRIMARY KEY, /* The directory */
	dir_type INT, /* Source (1) or include (2) or both (3)*/
	dir_order, /* Order in which the directories should be scanned for files */
	dir_attr INT/* Extra attributes saved as an integer of bitwise OR-ed values */
);

/*
** The FILES table contains the list of files to be parsed for creating the
** project database. A scanning function is provided which will read necessary
** information from the PARSERS table and populate the FILES table by scanning
** the project directory for files and matching their extensions with the ones
** in the PARSERS table. You should use your build system to add/remove items
** from this list so that files that are not needed are removed. You can also
** change the parser associated with individual files.
*/
create table FILES (
	file_id INTEGER PRIMARY KEY, /* Unique identifier for this file */
	file_name VARCHAR(256), /* Name of the file */
	file_dir VARCHAR(768), /* Directory where the file resides */
	file_path VARCHAR(1024), /* name + dir - saved for convenience */
	parser_id INTEGER, /* ID of the parser to be used with this file */
	parse_time INTEGER, /* Unix timestamp indicating when this file was last parsed */
	file_attr INTEGER /* Extra attributes of the file (bitwise OR-ed values) */
  /* Not supported by SQLite */
  /* FOREIGN KEY parser_id REFERENCES PARSERS(parser_id) */
);

create unique index FILES_UK1 on FILES (file_path);

/*
** The DEPENDENCIES table tracks interdependencies between files.
** The dependencies can be local or external to the project. External
** dependencies are normally scanned only once (if ever) for symbols.
*/
create table DEPENDENCIES (
  file_id INTEGER, /* Id of the parent file */
  inc_file_id INTEGER, /* Id of the included file */
  start_pos, /* Position in file_id where inc_file_id is included */
  PRIMARY KEY (file_id, inc_file_id)
  /* Not supported by SQLITE */
  /* FOREIGN KEY file_id REFERENCES FILES(file_id)
  FOREIGN KEY inc_file_id REFERENCES FILES(file_id) */
);

/* This is not needed/supported with sqlite */
/*CREATE SEQUENCE SYMBOLS_sym_id_seq;*/
/*
** The SYMBOLS table is the most important table and contains all information
** about the project symbols except cross-reference, comments and inheritence.
** There are a number of indexes on this table to facilitate faster searches.
*/
create table SYMBOLS (
  /* This is not supported by sqlite*/
  /*sym_id	INTEGER DEFAULT nextval ('"SYMBOLS_sym_id_seq"'::text) PRIMARY KEY, */
  /* Quoting from SQLite docs:
  Beginning with version SQLite 2.3.4, If you declare a column of
  a table to be INTEGER PRIMARY KEY, then whenever you insert a NULL
  into that column of the table, the NULL is automatically converted
  into an integer which is one greater than the largest value of that
  column over all other rows in the table, or 1 if the table is empty.*/
  sym_id INTEGER PRIMARY KEY,
  sym_name VARCHAR(1024), /* Name of the symbol */
  sym_type VARCHAR(64), /* Symbol type - see TYPES.sym_type */
  sym_attr INTEGER, /* Extra attributes lof this symbol - bitwise ORed */
  parent_sym VARCHAR(1024), /* Symbol this is a child of, e.g. class name for class members */
  file_id INTEGER, /* Id of the file where this symbol is found */
  start_line INTEGER, /* Starting line number of the symbol */
  start_col INTEGER, /* Starting column of the symbol */
  end_line INTEGER, /* Endinf line number of the symbol */
  end_col INTEGER, /* Ending column of the symbol */
  dec_type VARCHAR(64), /* Declaration or return type */
  template_class VARCHAR(1024), /* Template class for template declarations, etc. */
  arg_types VARCHAR(1024), /* Comma seperated list of argument types */
  arg_names VARCHAR(1024) /* Comma seperated list of argument names */
  /* These are not supported by SQLite */
  /*FOREIGN KEY(sym_type) REFERENCES types(sym_type)*/
  /*FOREIGN KEY(file_id) REFERENCES files(file_id) */
);

create index SYMBOLS_IDX1 on SYMBOLS (sym_name);
create index SYMBOLS_IDX2 on SYMBOLS (file_id);
create index SYMBOLS_IDX3 on SYMBOLS (sym_type);
create index SYMBOLS_IDX4 on SYMBOLS (parent_sym);

/*
** The XREF table shows cross-reference information between symbols inside
** the project. It is only populated on request since it can potentially
** bloat the size of the data file significantly.
*/
create table XREF (
  parent_sym /* Name of the parent symbol (class) where the xref occurs */
, sym_name /* Name of the symbol where the xref occurs */
, sym_type /* Type of the symbol where xref occurs */
, arg_types /* Argument types of the function where this xref occurs */
, ref_parent_sym /* Parent symbol of the xref-ed symbol */
, ref_sym_name /* Name of the referenced symbol */
, ref_sym_type /* Type of the referenced symbol */
, ref_arg_types /* Argument types of the referenced symbol */
, access_type /* Access type of the reference */
, file_id /* Id of the file where this xref occurs */
, start_pos /* Starting position in the file where the xref occurs */
);

create index XREF_IDX1 on XREF (parent_sym, sym_name, sym_type);
create index XREF_IDX2 on XREF (ref_parent_sym, ref_sym_name, ref_sym_type);
create index XREF_IDX3 on XREF (file_id);

/*
** The INHERITANCE table contains symbol hierarchy information. This is stored
** in a seperate table for quicker access.
*/
create table INHERITANCE (
  sym_name /* Name of the symbol (class) */
, parent_sym /* Name of the parent symbol */
, file_id /* File where this inheritance information si retreived from */
, start_line /* Starting line */
, start_col /* Starting column */
, end_line /* End line */
, end_col /* End column */
, attr /* Attributes */
, class_template /* Class template */
, inherit_template /* Inherited class template */
);

create index INHERITANCE_IDX1 on INHERITANCE (sym_name);
create index INHERITANCE_IDX2 on INHERITANCE (parent_sym);

/*
** The LOCALS table contains information about local variables. This is stored
** seperately since this is optional and bloats the SYMBOLS table too much
** when turned on.
*/
create table LOCALS (
  sym_name /* Name of the symbol where the local variable is declared */
, var_name /* Name of the local variable */
, file_id /* File where this local is declared */
, start_line /* Starting line number */
, start_col /* Starting column */
, end_line /* Ending line number */
, end_col /* Ending column */
, attr /* Extra attributes */
, var_type /* Type fo variable */
);

create index LOCALS_IDX1 on LOCALS (sym_name);
create index LOCALS_IDX2 on LOCALS (file_id);

/*
** The COMMENTS table (optionally) parsers and stores comments which can be used
** to generate doxygen-like documentation automatically with help from other
** tables.
*/
create table COMMENTS (
	sym_name, /* Name of the symbol associated with the comment */
	parent_sym, /* Parent symbol for the symbol associated with the comment */
	file_id, /* File where the comment occured */
	start_line, /* Line where the comment started */
	start_col, /* Column where the comment started */
	comment /* The comment itself */
);

create INDEX COMMENTS_IDX1 on COMMENTS(sym_name, parent_sym);
create INDEX COMMENTS_IDX2 on COMMENTS(file_id);
